#include <Adafruit_MAX31865.h>
#include <DueTimer.h>
#include <math.h>

// Pas de temps de la boucle, en secondes
#define TIME_STEP_LOOP 2

float time_since_start = 0; // s

/*
 * Prise en charge des deux capteurs PT100 
 * On appelle la méthode temperature(PT100_RNOMINAL,MAX31865_RREF) pour en lire les valeurs 
 * Les arguments sont les numéros de pins des bus I2C.
 */
#define MAX31865_RREF 430.0
#define PT100_RNOMINAL 100.0
Adafruit_MAX31865 rtdairsock = Adafruit_MAX31865(30,31,32,33);
Adafruit_MAX31865 rtdfinal = Adafruit_MAX31865(50,51,52,53);

/*
 * Utilisation du DAC de l'Arduino Due pour commander le servomoteur 0-10V
 */

// Tension maximale de DAC+ampli (dépend du gain de l'ampli externe)
#define VOLTAGE_AT_MAX_DAC 7.01

bool write_DAC_valve (float servo_voltage) {
  bool saturated = false;

  if (servo_voltage > VOLTAGE_AT_MAX_DAC){
    servo_voltage = VOLTAGE_AT_MAX_DAC;
    saturated = true;
  }
  else if (servo_voltage < 0.) {
    servo_voltage = 0.;
    saturated = true;
  }

  analogWrite(DAC0, (int)( servo_voltage / VOLTAGE_AT_MAX_DAC * 4095 ));

  return saturated;
}

/*
 * Le 1er étage désigne la boucle de régulation "interne" 
 * Régule la température en sortie de chaussette
 * Toutes les variables et fonctions sont précédées de pr.
 * 
 * Le 2eme étage désigne la boucle de régulation "externe"
 * Régule la température sur la table de sciences
 * Toutes les variables et fonctions sont précédées de de.
 */

#define ERROR_DELTA_HIST_SIZE 40

namespace InnerLoop {
  float setpoint = 0; // °C chaussettes
  float measurement_accumulator = 0;
  unsigned int measurement_accumulated_samples = 0;
  float measurement; // feedback, °C

  float error = 0; // °C
  float error_int = 0; // s·°C
  float error_delta_hist[ERROR_DELTA_HIST_SIZE]; // buffer storing consecutive deltas of the error, for error_deriv to be an average over ERROR_DELTA_HIST_SIZE steps
  float error_deriv = 0; // °C/s

  // PID coeffs (parallel structure + global gain)
  const float Kg = -1/1.34; // 1/°C
//  const float Kp = 2.5;
//  const float Ti = 250; // s
//  const float Td = 100; // s
  const float Kp = 3;
  const float Ti = 150; // s
  const float Td = 100; // s

  // Point de fonctionnement de la vanne
  float oppoint_valve = 1.0; // V

  // attempt to linearize the response of the valve : applying a voltage with a curve `a·cmd^expon` with expon>1
  // => higher gain at high voltage (valve nearly open), lower gain at low voltage (valve nearly closed)
  float cmd_to_valve_servo_voltage (float cmd) {
    constexpr float a = 0.5; // V
    constexpr float expon = 2.0;
    return a * powf(max( 0.f, powf(oppoint_valve/a, 1/expon) + cmd ), expon);
  }
  float cmd_valve; // V
}

namespace OuterLoop {
  float setpoint = 22.0; // °C pièce
  float measurement_accumulator = 0;
  unsigned int measurement_accumulated_samples = 0;
  float measurement; // feedback, °C

  float error = 0; // °C
  float error_int = 0; // s·°C
  float error_delta_hist[ERROR_DELTA_HIST_SIZE]; // buffer storing consecutive deltas of the error
  float error_deriv = 0; // °C/s

  // PID coeffs (parallel structure)
  const float Kp = 0.4;
  const float Ti = 400; // s
  const float Td = 50; // s
  // implicit global gain of -1 °C_int/°C_out

  // Point de fonctionnement pour la température des chaussettes d'air
  float oppoint_inner_setpoint = 21; //°C
}

float compute_average_error_delta (const float* error_delta_hist) {
  float sum = 0; 
  for (unsigned short i = 0; i < ERROR_DELTA_HIST_SIZE; i++) 
     sum += error_delta_hist[i];
  return sum / ERROR_DELTA_HIST_SIZE;
}

// error_delta_hist are used as circular buffers, error_delta_hist_current_position is the index
unsigned short error_delta_hist_current_position = 0;

/*
 * Loop timer
 */

DueTimer loop_timer = DueTimer(6); // let's use timer 6 because why not

volatile bool do_time_step = false; // flag which is raised each TIME_STEP_LOOP seconds

void interrupt_handler () {
  do_time_step = true;
}

/*
 * Initialisation
 */

void setup() {
  // initialisation des deux capteurs, en mode 4 fils 
  rtdairsock.begin(MAX31865_4WIRE);
  rtdfinal.begin(MAX31865_4WIRE);

  // résolution utilisée pour le DAC
  analogWriteResolution(12);

  // port série utilisé pour le écrire le log
  Serial.begin(115200);

  // dérivée nulle au début
  for (unsigned short i = 0; i < ERROR_DELTA_HIST_SIZE; i++) {
    InnerLoop::error_delta_hist[i] = 0;
    OuterLoop::error_delta_hist[i] = 0;
  }

  // set up the loop timer
  loop_timer.attachInterrupt(interrupt_handler);
  loop_timer.setFrequency(1.0/TIME_STEP_LOOP);
  loop_timer.start();
}

/*
 * Boucles d'asservissement.
 */

void loop() {

  /*** Accumulation of measurements during dead time to improve resolution ***/

  while (not do_time_step) {
    // loop will be stopped by the timer
    
    uint8_t fault;
    float temp;
    
    temp = rtdairsock.temperature(PT100_RNOMINAL,MAX31865_RREF);
    fault = rtdairsock.readFault();
    if (not fault) {
      InnerLoop::measurement_accumulator += temp;
      InnerLoop::measurement_accumulated_samples++;
    }
  
    temp = rtdfinal.temperature(PT100_RNOMINAL,MAX31865_RREF);
    fault = rtdfinal.readFault();
    if (not fault) {
      OuterLoop::measurement_accumulator += temp;
      OuterLoop::measurement_accumulated_samples++;
    }
  
    delay(10);
  }
  
  /*** Étage interne ***/

  {
    using namespace InnerLoop;

    measurement = measurement_accumulator / measurement_accumulated_samples;
    measurement_accumulator = 0; // reset the accumulator
    measurement_accumulated_samples = 0;
    
    float last_error = error;
    error = setpoint - measurement;

    float last_error_int = error_int;
    error_int += error * TIME_STEP_LOOP;
    
    error_delta_hist[error_delta_hist_current_position] = error - last_error;
    error_deriv = compute_average_error_delta(error_delta_hist) / TIME_STEP_LOOP;

    // summation of P, I, D and computation of command voltage
    float cmd = Kg * ( Kp * error + 1/Ti * error_int + Td * error_deriv );
    cmd_valve = cmd_to_valve_servo_voltage(cmd);

    // smooth startup to avoid a dramatic change of temperature when the loops are started
    float r = time_since_start / (Ti / 2.);
    if (r < 1)
      cmd_valve = r * cmd_valve + (1-r) * oppoint_valve;

    bool saturated = write_DAC_valve( cmd_valve );

    // arrêt de l'intégration si la vanne est entièrement ouverte ou fermée
    if (saturated)
      error_int = last_error_int;
  }

  /*** Étage externe ***/

  {
    using namespace OuterLoop;

    measurement = measurement_accumulator / measurement_accumulated_samples;
    measurement_accumulator = 0; // reset the accumulator
    measurement_accumulated_samples = 0;

    float last_error = error;
    error = setpoint - measurement;

    float last_error_int = error_int;
    error_int += error * TIME_STEP_LOOP;

    error_delta_hist[error_delta_hist_current_position] = error - last_error;
    error_deriv = compute_average_error_delta(error_delta_hist) / TIME_STEP_LOOP;

    float cmd = Kp * error + 1/Ti * error_int + Td * error_deriv;

    InnerLoop::setpoint = oppoint_inner_setpoint + cmd;

    // arrêt de l'intégration si la température de consigne n'est pas raisonnable
    if (InnerLoop::setpoint > OuterLoop::setpoint) {
      InnerLoop::setpoint = OuterLoop::setpoint;
      error_int = last_error_int;
    }
    else if (InnerLoop::setpoint < 17) {
      InnerLoop::setpoint = 17;
      error_int = last_error_int;
    }
  }

  error_delta_hist_current_position = (error_delta_hist_current_position + 1) % ERROR_DELTA_HIST_SIZE;

  /*** Log ***/

  String pt_tsock = "TI:" + String(InnerLoop::measurement, 3);
  String pt_ttable = "TO:" + String(OuterLoop::measurement, 3);
  
  String pt_cmd = "CMD:" + String(InnerLoop::cmd_valve, 3);
  String pt_consigne = "SETI:" + String(InnerLoop::setpoint, 3);

  Serial.println(pt_tsock + ";" + pt_ttable + ";" + pt_cmd + ";" + pt_consigne);

  time_since_start += TIME_STEP_LOOP;
  do_time_step = false; // reset the flag
}
